/**
 * An implementation of various Load Balancing schemes, to examine the concepts of Isoefficiency.
 */


// Import headers.
#include <ctype.h>
#include <limits.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/sysinfo.h>
#include <unistd.h>


// Method Declaration.
void validate_args();       // Validates passed user args.
void display_max_types();   // Displays a few data type max values.
void run_program();         // Main program logic. Handles various Load Balance schemas.
void simulate_load();       // Simulates running one load.


// Global Variables.
int thread_count;
int seconds_per_index;
int indexes_per_load;
int total_loads;


/**
 * Program's main. Initialized and runs program.
 */
int main(int argc, char* argv[]) {
    printf("Initializing program.\n");
    printf("\n");

    // Validate program args.
    validate_args(argc, argv);
    // If we got this far, then args are valid. Proceed with program.
    int total_processors = get_nprocs();

    // Display args.
    printf("Provided values:\n");
    printf("    %i Total Loads\n", total_loads);
    printf("    %i Indexes Per Load\n", indexes_per_load);
    printf("    %i Seconds Per Index\n", seconds_per_index);
    printf("\n");
    printf("    Total Expected work = TotalLoads * IndexesPerLoad * SecondsPerIndex\n");
    printf("                        = %i * %i * %i\n", total_loads, indexes_per_load, seconds_per_index);
    int seconds_active_work = total_loads * indexes_per_load * seconds_per_index;
    printf("                        = %i Seconds Active Work\n", seconds_active_work);
    printf("                        = %i Minutes Active Work\n", (seconds_active_work / 60));
    printf("\n");
    printf("    %i Total System Processors\n", total_processors);
    printf("\n");
    printf("    Expected total execution time with ideal processor usage:\n");
    printf("    (Work / Processor Count) = %i Minutes\n", (seconds_active_work / 60 / total_processors));
    printf("\n");

    // Run program.
    display_max_types();
    run_program(total_processors, argv);

    printf("\n");
    printf("Terminating program.\n");
    exit(0);
}


/**
 * Validate program args.
 */
void validate_args(int argc, char* argv[]) {

    if (argc == 4) {
        // Expected number of params. Validate passed args.

        // Validate "seconds_per_index" value. Should be between 1 and 10.
        seconds_per_index = strtol(argv[1], NULL, 10);
        if ((seconds_per_index < 1) || (seconds_per_index > 10)) {
            printf("Arg1 (seconds_per_index) should be int between 1 and 10.\n");
            printf("\n");
            printf("Terminating program.\n");
            exit(1);
        }

        // Validate "indexes_per_load" value. Should be between 100 and 10,000.
        indexes_per_load = strtol(argv[2], NULL, 10);
        if ((indexes_per_load < 100) || (indexes_per_load > 10000)) {
            printf("Arg2 (indexes_per_load) should be int between 100 and 10,000.\n");
            printf("\n");
            printf("Terminating program.\n");
            exit(1);
        }

        // Validate "total_loads" value. Should be between 100 and 10,000.
        total_loads = strtol(argv[3], NULL, 10);
        if ((total_loads < 100) || (total_loads > 10000)) {
            printf("Arg3 (total_loads) should be int between 100 and 10,000.\n");
            printf("\n");
            printf("Terminating program.\n");
            exit(1);
        }

    } else if (argc > 4) {
        // Too many args. Error.
        printf("Too many args passed. Got %i. Expected 3.\n", (argc - 1));
        printf("\n");
        printf("Terminating program.\n");
        exit(1);
    } else {
        // Too few args. Error.
        printf("Too few args passed. Got %i. Expected 3.\n", (argc - 1));
        printf("\n");
        printf("Terminating program.\n");
        exit(1);
    }
}


/**
 * Helper function to display some max values for system.
 */
void display_max_types() {
    printf("\n\n");

    printf("Int max is %i.\n", INT_MAX);
    printf("Unsigned Int max is %u.\n", UINT_MAX);
    printf("Long Max is %ld.\n", LONG_MAX);
    printf("Unsigned Long max is %lu.\n", ULONG_MAX);

    printf("\n\n");
}


/**
 * Main program logic.
 * Entry point for running various Load Balancing schemes.
 */
void run_program(int total_processors, char* argv[]) {
    printf("Running core program logic.");
    printf("\n");

    // Test load logic with no parallelization. Debugging only.
    // simulate_load("main");

    printf("\n");
    printf("Core program logic complete.");
}


/**
 * Function to simulate working on a load.
 */
void simulate_load(char* thread_name) {
    printf("Simulating load processing on thread \"%s\".\n", thread_name);

    int load_counter = 0;
    int index = 0;

    // Iterate through loads. Each load has indexes to simulate "work".
    while (load_counter < total_loads) {

        // Increment load num.
        load_counter += 1;
        printf("Simulating load %i on thread \"%s\".\n", load_counter, thread_name);

        // Simulate running the load.
        index = 0;
        while (index < indexes_per_load) {

            // Print out index value.
            if ((index % 10) == 0) {
                printf("load %i / %i    index %i / %i\n", load_counter, total_loads, index, indexes_per_load);
            }

            // Wait 1 second to simulate work.
            sleep(1);

            // Increment index.
            index += 1;
        }

    }

}
